@@ -28,17 +28,18 @@ from typing import (
 )
 
 import redis
+from redis.client import Pipeline
 from pydantic import BaseModel, validator
 from pydantic.fields import FieldInfo as PydanticFieldInfo
 from pydantic.fields import ModelField, Undefined, UndefinedType
 from pydantic.main import ModelMetaclass, validate_model
 from pydantic.typing import NoArgAnyCallable
 from pydantic.utils import Representation
-from redis.client import Pipeline
 from ulid import ULID
 
 from ..checks import has_redis_json, has_redisearch
 from ..connections import get_redis_connection
+from ..unasync_util import ASYNC_MODE
 from .encoders import jsonable_encoder
 from .render_tree import render_tree
 from .token_escaper import TokenEscaper
@@ -116,6 +117,17 @@ def is_supported_container_type(typ: Optional[type]) -> bool:
 
 def validate_model_fields(model: Type["RedisModel"], field_values: Dict[str, Any]):
     for field_name in field_values.keys():
+        if "__" in field_name:
+            obj = model
+            for sub_field in field_name.split("__"):
+                if not hasattr(obj, sub_field):
+                    raise QuerySyntaxError(
+                        f"The update path {field_name} contains a field that does not "
+                        f"exit on {model.__name__}. The field is: {sub_field}"
+                    )
+                obj = getattr(obj, sub_field)
+            return
+
         if field_name not in model.__fields__:
             raise QuerySyntaxError(
                 f"The field {field_name} does not exist on the model {model.__name__}"
@@ -136,6 +148,23 @@ def decode_redis_value(
         return obj.decode(encoding)
 
 
+class PipelineError(Exception):
+    """A Redis pipeline error."""
+
+
+def verify_pipeline_response(
+    response: List[Union[bytes, str]], expected_responses: int = 0
+):
+    # TODO: More generic pipeline verification here (what else is possible?),
+    #  plus hash and JSON-specific verifications in separate functions.
+    actual_responses = len(response)
+    if actual_responses != expected_responses:
+        raise PipelineError(
+            f"We expected {expected_responses}, but the Redis "
+            f"pipeline returned {actual_responses} responses."
+        )
+
+
 @dataclasses.dataclass
 class NegatedExpression:
     """A negated Expression object.
@@ -744,6 +773,7 @@ class FindQuery:
         validate_model_fields(self.model, field_values)
         pipeline = self.model.db().pipeline() if use_transaction else None
 
+        # TODO: async for here?
         for model in self.all():
             for field, value in field_values.items():
                 setattr(model, field, value)
@@ -783,12 +813,41 @@ class FindQuery:
                that result, then we should clone the current query and
                give it a new offset and limit: offset=n, limit=1.
         """
+        if ASYNC_MODE:
+            raise QuerySyntaxError(
+                "Cannot use [] notation with async code. "
+                "Use FindQuery.get_item() instead."
+            )
         if self._model_cache and len(self._model_cache) >= item:
             return self._model_cache[item]
 
         query = self.copy(offset=item, limit=1)
 
-        return query.execute()[0]
+        return query.execute()[0]  # noqa
+
+    def get_item(self, item: int):
+        """
+        Given this code:
+            await Model.find().get_item(1000)
+
+        We should return only the 1000th result.
+
+            1. If the result is loaded in the query cache for this query,
+               we can return it directly from the cache.
+
+            2. If the query cache does not have enough elements to return
+               that result, then we should clone the current query and
+               give it a new offset and limit: offset=n, limit=1.
+
+        NOTE: This method is included specifically for async users, who
+        cannot use the notation Model.find()[1000].
+        """
+        if self._model_cache and len(self._model_cache) >= item:
+            return self._model_cache[item]
+
+        query = self.copy(offset=item, limit=1)
+        result = query.execute()
+        return result[0]
 
 
 class PrimaryKeyCreator(abc.ABC):
@@ -908,7 +967,7 @@ class PrimaryKey:
     field: ModelField
 
 
-class MetaProtocol(Protocol):
+class BaseMeta(abc.ABC):
     global_key_prefix: str
     model_key_prefix: str
     primary_key_pattern: str
@@ -942,7 +1001,7 @@ class DefaultMeta:
 
 
 class ModelMeta(ModelMetaclass):
-    _meta: MetaProtocol
+    _meta: BaseMeta
 
     def __new__(cls, name, bases, attrs, **kwargs):  # noqa C901
         meta = attrs.pop("Meta", None)
@@ -1056,6 +1115,10 @@ class RedisModel(BaseModel, abc.ABC, metaclass=ModelMeta):
     def delete(self):
         return self.db().delete(self.key())
 
+    @classmethod
+    def get(cls, pk: Any) -> "RedisModel":
+        raise NotImplementedError
+
     def update(self, **field_values):
         """Update this model instance with the specified key-value pairs."""
         raise NotImplementedError
@@ -1162,9 +1225,34 @@ class RedisModel(BaseModel, abc.ABC, metaclass=ModelMeta):
         return d
 
     @classmethod
-    def add(cls, models: Sequence["RedisModel"]) -> Sequence["RedisModel"]:
-        # TODO: Add transaction support
-        return [model.save() for model in models]
+    def add(
+        cls,
+        models: Sequence["RedisModel"],
+        pipeline: Optional[Pipeline] = None,
+        pipeline_verifier: Callable[..., Any] = verify_pipeline_response,
+    ) -> Sequence["RedisModel"]:
+        if pipeline is None:
+            # By default, send commands in a pipeline. Saving each model will
+            # be atomic, but Redis may process other commands in between
+            # these saves.
+            db = cls.db().pipeline(transaction=False)
+        else:
+            # If the user gave us a pipeline, add our commands to that. The user
+            # will be responsible for executing the pipeline after they've accumulated
+            # the commands they want to send.
+            db = pipeline
+
+        for model in models:
+            # save() just returns the model, we don't need that here.
+            model.save(pipeline=db)
+
+        # If the user didn't give us a pipeline, then we need to execute
+        # the one we just created.
+        if pipeline is None:
+            result = db.execute()
+            pipeline_verifier(result, expected_responses=len(models))
+
+        return models
 
     @classmethod
     def redisearch_schema(cls):
@@ -1201,11 +1289,12 @@ class HashModel(RedisModel, abc.ABC):
         else:
             db = pipeline
         document = jsonable_encoder(self.dict())
+        # TODO: Wrap any Redis response errors in a custom exception?
         db.hset(self.key(), mapping=document)
         return self
 
     @classmethod
-    def all_pks(cls):
+    def all_pks(cls):  # type: ignore
         key_prefix = cls.make_key(cls._meta.primary_key_pattern.format(pk=""))
         # TODO: We assume the key ends with the default separator, ":" -- when
         #  we make the separator configurable, we need to update this as well.
@@ -1214,7 +1303,9 @@ class HashModel(RedisModel, abc.ABC):
         # TODO: Also, we need to decide how we want to handle the lack of
         #  decode_responses=True...
         return (
-            key.split(":")[-1] if isinstance(key, str) else key.decode(cls.Meta.encoding).split(":")[-1]
+            key.split(":")[-1]
+            if isinstance(key, str)
+            else key.decode(cls.Meta.encoding).split(":")[-1]
             for key in cls.db().scan_iter(f"{key_prefix}*", _type="HASH")
         )
 
@@ -1345,13 +1436,16 @@ class HashModel(RedisModel, abc.ABC):
 
 class JsonModel(RedisModel, abc.ABC):
     def __init_subclass__(cls, **kwargs):
-        if not has_redis_json(cls.db()):
+        # Generate the RediSearch schema once to validate fields.
+        cls.redisearch_schema()
+
+    def __init__(self, *args, **kwargs):
+        if not has_redis_json(self.db()):
             log.error(
                 "Your Redis instance does not have the RedisJson module "
                 "loaded. JsonModel depends on RedisJson."
             )
-        # Generate the RediSearch schema once to validate fields.
-        cls.redisearch_schema()
+        super().__init__(*args, **kwargs)
 
     def save(self, pipeline: Optional[Pipeline] = None) -> "JsonModel":
         self.check()
@@ -1359,14 +1453,32 @@ class JsonModel(RedisModel, abc.ABC):
             db = self.db()
         else:
             db = pipeline
+        # TODO: Wrap response errors in a custom exception?
         db.execute_command("JSON.SET", self.key(), ".", self.json())
         return self
 
     def update(self, **field_values):
-        # TODO: Better support for embedded field models.
         validate_model_fields(self.__class__, field_values)
         for field, value in field_values.items():
-            setattr(self, field, value)
+            # Handle the simple update case first, e.g. city="Happy Valley"
+            if "__" not in field:
+                setattr(self, field, value)
+                continue
+
+            # Handle the nested update field name case, e.g. address__city="Happy Valley"
+            obj = self
+            parts = field.split("__")
+            path_to_field = parts[:-1]
+            target_field = parts[-1]
+
+            # Get the final object in a nested update field name, e.g. for
+            # the string address__city, we want to get self.address.city
+            for sub_field in path_to_field:
+                obj = getattr(obj, sub_field)
+
+            # Set the target field (the last "part" of the nested update
+            # field name) to the target value.
+            setattr(obj, target_field, value)
         self.save()
 
     @classmethod
